"""
Created on Tue Jan 8 10:36:04 2019

@author: marcusfalls

Input : Directory of NetCDF files from the coriolis floats individual VERTICAL profiles for multiple variables
Output: Directory of NetCDF files of different variables
"""

from netCDF4 import Dataset
import numpy as np
import os
import datetime
from datetime import datetime as DATETIME

print(datetime.datetime.now().strftime("%Y-%m-%d_%H:%M:%S"))
print('modules loaded')

directory='esarchive'  # choose 'local' (testing) or 'esarchive'

if directory=='local':
    #local
    inpath='/esarchive/scratch/mgalitap/ORCAS_FLOATS/floats_explore/coriolis/indiv_cycles/' # path to the netcdf files
    outpath='/home/Earth/mfalls/bgc-argo/testfiles' # place to store the output files
    files=['BR6901480_001D.nc','BR6901486_001D.nc','R6901480_001D.nc','R6901486_001D.nc','BR6901472_003.nc','R6901472_003.nc','BR6901480_156.nc','BR6901480_380.nc']

elif directory=='esarchive':
    #power9
    inpath='/esarchive/obs/argo/bgc/original_files/ftp.ifremer.fr/dac/'
    outpath='/esarchive/obs/argo/bgc/daily'
    #outpath = '/gpfs/scratch/bsc32/bsc32786/bgcargo/testfiles'

# extraction of variables from files
measurements=['BBP700','CDOM','CHLA','DOWNWELLING_PAR','TEMP','PSAL','NITRATE','DOXY']


def extract(path,file,varnames): # returns a dictionary of dictionaries of each variable in array format, also in the
                                 # dictionary is: 8 digit date, float no., latitude, longitude, pressure
    # path = string of directory file is in
    # file = filename (string)
    # varnames = list of strings of the variables that are to be stored

    def ninestonan(A): # in all float arrays, turns 99999.0 into NaN

        inds = np.where(A == 99999)
        #print('A',inds)
        for i in range(len(inds[0])):
            s = []
            for j in range(len(inds)):
                s.append(inds[j][i])
            s = tuple(s)
            A[s] = np.nan
        return A

    def emptytozero(A): # in all qc flag arrays, turns ' ' into '0' and converts everything to integer

        if isinstance(A[0], np.bytes_):
            B=np.tile(' ',np.shape(A))
            for i in range(len(A)):
                    B[i]=A[i].decode('UTF-8')
            A=B
        inds = np.where(A == ' ')
        #print('A',inds)
        for i in range(len(inds[0])):
            s = []
            for j in range(len(inds)):
                s.append(inds[j][i])
            s = tuple(s)
            A[s] = '0'
        #print(A)
        A.astype(int)
        return A

    def ordtostr(dt):

        stdate = DATETIME(1950, 1, 1, 0, 0, 0)
        stdate = DATETIME.toordinal(stdate)
        actdate = stdate + dt
        actdate = DATETIME.fromordinal(actdate)
        actdate = "{:04d}".format(actdate.year) + "{:02d}".format(actdate.month) + "{:02d}".format(
            actdate.day) + "{:02d}".format(actdate.hour) + "{:02d}".format(actdate.minute) + "{:02d}".format(
            actdate.second)

        return actdate

    data=Dataset(path+file,'r')
    keys = data.variables.keys()
    result = {}

    for var in varnames:

        ## obtaining variables

        newdict = {}
        if var in keys:
            newdict[var] = np.array(data.variables[var])
            newdict[var] = ninestonan(newdict[var])
            #determine which depth axis the variable is on
            newdict['INDEX']=0
            for i in range(len(newdict[var][:,0])):

                if not np.isnan(newdict[var][i,0]):
                    newdict['INDEX']=i
            ind=newdict['INDEX']
            print(ind)
            newdict[var] = newdict[var][ind]

            if var + '_QC' in keys:
                newdict[var + '_QC'] = np.array(data.variables[var + '_QC'])
                newdict[var + '_QC'] = emptytozero(newdict[var + '_QC'][ind])
            if var + '_ADJUSTED' in keys:
                newdict[var + '_ADJUSTED'] = np.array(data.variables[var + '_ADJUSTED'])
                newdict[var + '_ADJUSTED'] = ninestonan(newdict[var + '_ADJUSTED'][ind])
            if var + '_ADJUSTED_QC' in keys:
                newdict[var + '_ADJUSTED_QC'] = np.array(data.variables[var + '_ADJUSTED_QC'])
                newdict[var + '_ADJUSTED_QC'] = emptytozero(newdict[var + '_ADJUSTED_QC'][ind])
            if var + '_ADJUSTED_ERROR' in keys:
                newdict[var + '_ADJUSTED_ERROR'] = np.array(data.variables[var + '_ADJUSTED_ERROR'])
                newdict[var + '_ADJUSTED_ERROR'] = ninestonan(newdict[var + '_ADJUSTED_ERROR'][ind])
        if newdict !={}:
            result[var] = newdict

        ## return float number, date, time, pressure, LATITUDE AND LONGITUDE and add to dictionary too

        if type(data.variables['LATITUDE'][0]) == np.ma.core.MaskedConstant:
            print(file+' has no latitude')
            return None
        result['LATITUDE'] = float(data.variables['LATITUDE'][0])

        if type(data.variables['LONGITUDE'][0]) == np.ma.core.MaskedConstant:
            print(file+' has no longitude')
            return None
        result['LONGITUDE'] = float(data.variables['LONGITUDE'][0])

        result['PRES']=np.array(data.variables['PRES'])
        result['PRES']=ninestonan(result['PRES'])

        if type(data.variables['JULD'][0]) == np.ma.core.MaskedConstant:
            print(file+' has no timedate')
            return None
        result['JULD']=int(data.variables['JULD'][0])

        result['DATE_CREATION']=''.join([i.decode('UTF-8') for i in np.array(data.variables['DATE_CREATION'])])
        result['DATE_UPDATE']=''.join([i.decode('UTF-8') for i in np.array(data.variables['DATE_UPDATE'])])

        datetime = ordtostr(result['JULD'])
        MONTH = datetime[:6]
        DAY= datetime[6:8]
        TIME = datetime[8:]
        result['DAY']=int(DAY)
        result['MONTH']=MONTH
        result['TIME']=float(TIME)

        result['FLOAT_ID']=''.join([i.decode('UTF-8') for i in np.array(data.variables['PLATFORM_NUMBER'][0, :])])
        result['FLOAT_ID']=result['FLOAT_ID'].replace(' ','')

    #print(file + ' extracted')
    return result


def getattributes(path,files,varnames):
    # input: directory, files as a list, required variables
    # output: dictionary of the attributes of the varnames

    result={}
    allvarnames = varnames+['LATITUDE','LONGITUDE','PRES','JULD','DATE_CREATION','DATE_UPDATE']
    for var in allvarnames:
        for file in files:
            filedata = Dataset(path + file, 'r')
            keys = filedata.variables.keys()
            if var in keys:
                vardata=filedata[var]
                result[var] = {}
                if var in varnames:
                    endings=['','_QC','_ADJUSTED','_ADJUSTED_QC','_ADJUSTED_ERROR']
                    for end in endings:
                        result[var][var+end]={}
                        for attr in vardata.ncattrs():
                            if attr != '_FillValue':
                                result[var][var+end][attr] = getattr(vardata, attr)
                else:
                    for attr in vardata.ncattrs():
                        if attr!='_FillValue':
                            result[var][attr]=getattr(vardata, attr)
                break
        # go through the files until one contains that variable
    return result


def makenetcdf(extracteddata,varnames,attributes):

    # making a list of directories of variables to be made
    variables = []
    for file in extracteddata.keys():
        for var in extracteddata[file].keys():
            if var in varnames and var not in variables:
                variables.append(var)

    # making the netCDFs
    netcdfs = {}
    for var in variables:
        os.mkdir(var.lower() + '/')  # directory = ./$var/$var_YYYYMM
        os.chdir(var.lower() + '/')
        print('New subdirectory ' + var.lower() + ' created')

        # making a dictionary of all the keys of the extracted files containing the variable sorted by month
        # need to sort these in terms of day
        months = {}
        for oldfile in extracteddata.keys():
            if var in extracteddata[oldfile].keys():
                month = extracteddata[oldfile]['MONTH']
                day = extracteddata[oldfile]['DAY']
                if month in months:
                    if day in months[month].keys():
                        months[month][day].append(oldfile)
                    else:
                        months[month][day]= []
                        months[month][day].append(oldfile)
                else:
                    months[month]={day:[oldfile]}
        # Body of file creation here
        for month in months.keys(): # month=month/year in form yyyymm
            m=month[4:] # m=month in form mm
            y=month[:4] # y=year in form yyyy

            # create the netcdf and global attributes
            dataset = Dataset(var.lower() + '_' + month + '.nc', 'w', format='NETCDF4_CLASSIC')
            dataset.title=var.lower() +' dataset for '+ m +'-'+ y
            dataset.history='Created ' + datetime.datetime.now().strftime("%Y-%m-%d_%H:%M:%S")
            dataset.description="Argo float data aggregated based on time and variable, to fit the standards of the BSC Earth Science Department"
            dataset.references="http://www.argodatamgt.org/Documentation"
            dataset.sources='Argo floats maintained by the following institutions: aoml, bodc, coriolis, csio, csiro, incois, jma, kma, kordi, meds'
            print('Dataset ' + var.lower() + '_' + month + '.nc created')
            # taking the number of particular different floats for the month and their pressure data to determine the
            # dimensions of the netcdf

            FLOATS = []
            preslengths = []
            for day in months[month].keys():
                for file in months[month][day]:
                    floatid = extracteddata[file]['FLOAT_ID']
                    preslengths.append(np.shape(extracteddata[file]['PRES'])[1])
                    if floatid not in FLOATS:
                        FLOATS.append(floatid)

            # determine time dimension by length of month
            if m in ['01','03','05','07','08','10','12']:
                monthlen=31
            elif m in ['04','06','09','11']:
                monthlen=30
            elif m=='02':
                if int(y)%4==0:
                    monthlen=29
                else:
                    monthlen=28

            # create dimensions
            float = dataset.createDimension('float', len(FLOATS))  # number of different floats of interest - obtain from filenames
            pres = dataset.createDimension('pres', max(preslengths))  # the length of the longest 'PRES' array
            time = dataset.createDimension('time', monthlen)  # file per month with daily data - shorter months will have the last entries as NaN

            # creating the variables
            floats=dataset.createVariable('floats', np.int32, ('float',), fill_value=0)

            latitude = dataset.createVariable('latitude', np.float64, ('time','float'), fill_value=np.nan)
            latitude.setncatts(attributes['LATITUDE'])

            longitude = dataset.createVariable('longitude', np.float64, ('time','float'),  fill_value=np.nan)
            longitude.setncatts(attributes['LONGITUDE'])

            date_creation = dataset.createVariable('date_creation', np.int32, ('time','float'), fill_value=0)
            date_creation.setncatts(attributes['DATE_CREATION'])
            date_creation.long_name = "Date of original file's creation"

            date_update = dataset.createVariable('date_update', np.int32, ('time', 'float'), fill_value=0)
            date_update.setncatts(attributes['DATE_UPDATE'])
            date_update.long_name = "Date of original file's most recent update"

            pressure = dataset.createVariable('pressure', np.float32, ('time','float','pres'),  fill_value=np.nan)
            pressure.setncatts(attributes['PRES'])

            juld = dataset.createVariable('juld', np.float64, ('time',), fill_value=np.nan)
            juld.setncatts(attributes['JULD'])             # NEED TO WORK ON

            day = dataset.createVariable('day', np.int32, ('time',),  fill_value=0)
            day.standard_name = 'day of the month'
            day.units = 'day'

            VARS={}
            for d in range(1,monthlen+1):
                if d in months[month].keys():
                    for file in months[month][d]:
                        for key in extracteddata[file][var].keys():
                            if key.lower() not in VARS.keys() and key!='INDEX':
                                if 'QC' in key:
                                    VARS[key.lower()]=dataset.createVariable(key.lower(), np.int32, ('time', 'float','pres'), fill_value=0)
                                    VARS[key.lower()].setncatts(attributes[var][key])
                                else:
                                    VARS[key.lower()] = dataset.createVariable(key.lower(), np.float32, ('time', 'float', 'pres'),  fill_value=np.nan)
                                    VARS[key.lower()].setncatts(attributes[var][key])

            # putting data into the variables

            floats[:]=FLOATS

            for d in range(1,monthlen+1):
                if d in months[month].keys():
                    tindex = d-1
                    for file in months[month][d]:
                        data = extracteddata[file]
                        float = data['FLOAT_ID']

                        for fl in range(len(FLOATS)):
                            if FLOATS[fl]==float:
                                flindex=fl

                        lat=data['LATITUDE']
                        lon=data['LONGITUDE']
                        dc=data['DATE_CREATION']
                        du=data['DATE_UPDATE']

                        latitude[tindex,flindex] = lat
                        longitude[tindex,flindex] = lon
                        date_creation[tindex,flindex] = dc
                        date_update[tindex,flindex] = du

                        index=data[var]['INDEX']
                        pres=data['PRES'][index]
                        pressure[tindex, flindex, :len(pres)] = pres

                        JULD = data['JULD']
                        juld[tindex]=JULD
                        day[tindex]=d

                        vars = data[var]
                        for VAR in VARS.keys():
                            VARS[VAR][tindex,flindex,:len(pres)] = vars[VAR.upper()]

        # body of file creation ends
        os.chdir('..')  ##

#### EXTRACTION OF DATA AND ATTRIBUTES

## local
if directory=='local':
    alldata={}
    for i in files:
        data=extract(inpath,i,measurements)
        if data!=None:
            alldata['data_'+i]=data
    print(files, 'extracted')
    attrs = getattributes(inpath, files, measurements)

## esarchive
elif directory=='esarchive':
    alldata={}
    orgs=os.listdir(inpath)
    attrs={}
    for org in orgs:
        os.chdir(inpath + org)
        floatids=os.listdir(inpath + org)
        #floatids=floatids[:30]
        for floatid in floatids:
            if 'profiles' in os.listdir(inpath+org+'/'+floatid+'/'):
                os.chdir(inpath+org+'/'+floatid+'/profiles/')
                allfiles=os.listdir(inpath+org+'/'+floatid+'/profiles/')
                #allfiles=allfiles[:30]
                print('extracting ' + floatid)
                files=[]
                for i in allfiles:
                    if i[-4]!='D':
                        files.append(i)

                for f in files:
                    data=extract(inpath+org+'/'+floatid+'/profiles/', f, measurements)
                    if data!=None:
                        alldata['data_' + f] = data

                print(floatid + ' all extracted')
                attrsdir = getattributes(inpath + org + '/' + floatid + '/profiles/', files[:1], measurements) ## 07/02 changed to files[:1] so that it only checks the first file in each directory for the attributes
                for key in attrsdir.keys():
                    if key not in attrs.keys():
                        attrs[key]=attrsdir[key]

# CREATION OF DIRECTORY FOR OUTPUTS

# local
if directory=='local':
    os.chdir(outpath)
    newdir=datetime.datetime.now().strftime("%Y-%m-%d_%H:%M:%S")
    os.mkdir(newdir)
    print('New directory '+newdir+' created')
    os.chdir(newdir)

# esarchive
elif directory=='esarchive':
    os.chdir(outpath)

# CREATION OF NETCDF

makenetcdf(alldata, measurements, attrs)

print('Complete')